function [primal] = MRF_pairs(lcosts,dlab)
% An implementation of the Primal-Dual algorithm for Graph Cuts
% N. Komodakis and G. Tziritas, Approximate labeling via graph cuts based
% on linear programming, IEEE Transactions on Pattern Analysis and Machine Intelligence,
% Vol. 29, No. 8, 1436--1453,  2007.
%%
% In the array lcosts are given the values of the -log likelihoods for all
% image points. Its size is [number_rows,number_columns,number_labels]
%%
%   Only pairs of pixels are considered in 8-connection neighborhood
%   The weights of the pairs are equal, no data depending
%   (wcosts in MRF_pairs)
%%
% In the array dlab are given the distances between labels
% dlab is a symmetric matrix with zero diagonal, for example
% dlab = [0     1     1     1;
%         1     0     1     1;
%         1     1     0     3;
%         1     1     3     0];

[sizey,sizex,numlabels] = size(lcosts);
lcosts = single(lcosts);
dlab = single(dlab);

max_iters = 20;

lcosts = reshape(lcosts,sizex*sizey,numlabels);

%% Create the edges of a 2nd-order MRF

[x0,y0] = (meshgrid(single(1:sizex-1),single(1:sizey)));
hx = x0+1; hy = y0;
inds0 = single(sub2ind( [sizey sizex], y0(:), x0(:) ) - 1);
hinds = single(sub2ind( [sizey sizex], hy(:), hx(:) ) - 1);

[x0,y0] = (meshgrid(single(1:sizex),single(1:sizey-1)));
hx = x0; hy = y0+1;
inds1 = single(sub2ind( [sizey sizex], y0(:), x0(:) ) - 1);
vinds = single(sub2ind( [sizey sizex], hy(:), hx(:) ) - 1);

[x0,y0] = (meshgrid(single(1:sizex-1),single(1:sizey-1)));
hx = x0+1; hy = y0+1;
inds2 = single(sub2ind( [sizey sizex], y0(:), x0(:) ) - 1);
dinds = single(sub2ind( [sizey sizex], hy(:), hx(:) ) - 1);

[x0,y0] = (meshgrid(single(1:sizex-1),single(2:sizey)));
hx = x0+1; hy = y0-1;
inds3 = single(sub2ind( [sizey sizex], y0(:), x0(:) ) - 1);
ddinds = single(sub2ind( [sizey sizex], hy(:), hx(:) ) - 1);

clear hx hy x0 y0;

pairs = [inds0 hinds; inds1 vinds; inds2 dinds; inds3 ddinds];
clear inds0 hinds inds1 vinds inds2 dinds inds3 ddinds;
wcosts = single(ones(size(pairs,1), 1));

%% Run primal-dual algorithm
%
primal = PD_wrapper2( lcosts, pairs, wcosts, dlab, max_iters );
primal = reshape( primal, sizey, sizex )+1;
end